import ujson
import functools
from typing import Union
from inspect import signature, Parameter

from .contexts import context


def action(*args, verbs: Union[str, list, tuple]='any', encoding: str='utf-8',
           content_type: Union[str, None]=None, inner_decorator: Union[callable, None]=None, **kwargs):
    """
    Base action decorator

    :param verbs: Allowed HTTP methods
    :param encoding: Content encoding
    :param content_type: Content Type
    :param inner_decorator: Inner decorator
    """
    def decorator(func):

        if inner_decorator is not None:
            func = inner_decorator(func, *args, **kwargs)

        # Examining the signature, and counting the optional and positional arguments.
        positional_arguments, optional_arguments, keywordonly_arguments = [], [], []
        action_signature = signature(func)
        for name, parameter in action_signature.parameters.items():
            if name == 'self':
                continue

            if parameter.kind == Parameter.KEYWORD_ONLY:
                keywordonly_arguments.append((parameter.name, parameter.default))
            elif parameter.default is Parameter.empty:
                positional_arguments.append(parameter.name)
            else:
                optional_arguments.append((parameter.name, parameter.default))

        func.__nanohttp__ = dict(
            verbs=verbs,
            encoding=encoding,
            content_type=content_type,
            positional_arguments=positional_arguments,
            optional_arguments=optional_arguments,
            keywordonly_arguments=keywordonly_arguments,
            default_action='index'
        )

        return func

    if args and callable(args[0]):
        f = args[0]
        args = tuple()
        return decorator(f)

    return decorator


def jsonify(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if hasattr(result, 'to_dict'):
            result = result.to_dict()
        elif not isinstance(result, (list, dict, int, str)):
            raise ValueError('Cannot encode to json: %s' % type(result))

        return ujson.dumps(result, indent=4)

    return wrapper


#: HTML action decorator
html = functools.partial(action, content_type='text/html')

#: Plain Text action decorator
text = functools.partial(action, content_type='text/plain')

#: JSON action decorator
#:
#: accepts list, dict, int, str or objects have ``to_dict`` method.
json = functools.partial(action, content_type='application/json', inner_decorator=jsonify)

#: XML action decorator
xml = functools.partial(action, content_type='application/xml')

#: Binary-data action decorator
binary = functools.partial(action, content_type='application/octet-stream', encoding=None)


def ifmatch(tag: Union[str, int, callable]):
    """ Validate ``If-Match`` header with given tag argument """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            context.etag_match(tag() if callable(tag) else tag)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def etag(*args, tag: Union[str, int, callable, None]=None):
    """
    Validate ``If-None-Match`` and response with ``ETag`` header

    tag is getting from ``tag`` argument or response object ``__etag__`` attribute
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*a, **kw):
            _etag = tag() if callable(tag) else tag
            if _etag is not None:
                context.etag_none_match(_etag)
                return func(*a, **kw)
            else:
                result = func(*a, **kw)
                if hasattr(result, '__etag__'):
                    _etag = result.__etag__
                if _etag:
                    context.etag_none_match(_etag)
                return result
        return wrapper

    if args and callable(args[0]):
        return decorator(args[0])

    return decorator





from .exceptions import HttpStatus, HttpBadRequest, HttpUnauthorized, HttpForbidden, HttpNotFound, \
    HttpMethodNotAllowed, HttpConflict, HttpGone, HttpRedirect, HttpMovedPermanently, HttpFound, \
    HttpInternalServerError, HttpNotModified, HttpBadGatewayError
from .controllers import Controller, RestController, Static, RegexRouteController
from .decorators import action, html, json, xml, binary, text, ifmatch, etag
from .helpers import quickstart, LazyAttribute
from .cli import main
from .contexts import context, ContextIsNotInitializedError
from .configuration import settings, configure
from .application import Application
from .validation import validate

__version__ = '0.28.0'



import ujson
import traceback

from .contexts import context
from .configuration import settings


class HttpStatus(Exception):
    status = None

    def __init__(self, status=None):
        if status is not None:
            self.status = status

        super().__init__(self.status)

    def render(self):
        stack_trace = traceback.format_exc()
        if context.response_content_type == 'application/json':
            return ujson.encode(
                dict(
                    stackTrace=stack_trace) if settings.debug else dict()
            )
        else:
            context.response_encoding = 'utf-8'
            context.response_content_type = 'text/plain'
            return stack_trace if settings.debug else ''


class HttpBadRequest(HttpStatus):
    status = '400 Bad Request'


class HttpUnauthorized(HttpStatus):
    status = '401 Unauthorized'


class HttpForbidden(HttpStatus):
    status = '403 Forbidden'


class HttpNotFound(HttpStatus):
    status = '404 Not Found'


class HttpMethodNotAllowed(HttpStatus):
    status = '405 Method Not Allowed'


class HttpConflict(HttpStatus):
    status = '409 Conflict'


class HttpGone(HttpStatus):
    status = '410 Gone'


class HttpPreconditionFailed(HttpStatus):
    status = '412 Precondition Failed'


class HttpRedirect(HttpStatus):
    """
    This is an abstract class for all redirects.
    """

    def __init__(self, location, *args, **kw):
        context.response_headers.add_header('Location', location)
        super().__init__(*args, **kw)


class HttpMovedPermanently(HttpRedirect):
    status = '301 Moved Permanently'


class HttpFound(HttpRedirect):
    status = '302 Found'


class HttpNotModified(HttpStatus):
    status = '304 Not Modified'


class HttpInternalServerError(HttpStatus):
    status = '500 Internal Server Error'

    @property
    def info(self):
        if settings.debug:
            return traceback.format_exc()
        return 'Server got itself in trouble'


class HttpBadGatewayError(HttpStatus):
    status = '502 Bad Gateway'



import re
import functools

from nanohttp import context
from nanohttp.exceptions import HttpStatus, HttpBadRequest


class Field:
    def __init__(self, title, form=True, query_string=False, required=None,
                 type_=None, minimum=None, maximum=None, pattern=None,
                 min_length=None, max_length=None):
        self.title = title
        self.form = form
        self.query_string = query_string
        self.criteria = []

        if required:
            self.criteria.append(RequiredValidator(required))

        if type_:
            self.criteria.append(TypeValidator(type_))

        if minimum:
            self.criteria.append(MinimumValidator(minimum))

        if maximum:
            self.criteria.append(MaximumValidator(maximum))

        if pattern:
            self.criteria.append(PatternValidator(pattern))

        if min_length:
            self.criteria.append(MinLengthValidator(min_length))

        if max_length:
            self.criteria.append(MaxLengthValidator(max_length))

    def validate(self, container):
        for criterion in self.criteria:
            criterion.validate(self, container)

        return container


class Criterion:
    def __init__(self, expression):
        if isinstance(expression, tuple):
            error = expression[1]
            self.expression = expression[0]
        else:
            self.expression = expression
            error = '400 Bad request'

        parsed_error = error.split(' ', 1)

        self.status_code = int(parsed_error[0])

        if len(parsed_error) == 2:
            self.status_text = parsed_error[1]
        else:
            self.status_text = 'Bad request'

    def validate(self, field: Field, container: dict) -> None:
        value = container.get(field.title)
        if value is None:
            return
        container[field.title] = self._validate(value, container, field)

    def _validate(self, value, container: dict, field: Field
                  ):  # pragma: no cover
        """
        It must be overridden in the child class.

        This method should raise exception if the criterion is not met. there
        is a chance to set
        a new value because the
        container is available here.
        :param value: The value to validate
        :param field:
        :param container:
        :return:
        """
        raise NotImplementedError()

    def create_exception(self):
        if self.status_code == 400:
            return HttpBadRequest

        return HttpStatus(
            status=f'{self.status_code} {self.status_text}'
        )


# noinspection PyAbstractClass
class RequiredValidator(Criterion):

    def validate(self, field, container):
        if field.title not in container:
            raise self.create_exception()


class TypeValidator(Criterion):

    def _validate(self, value, container, field):
        type_ = self.expression
        try:
            return type_(value)
        except (ValueError, TypeError):
            raise self.create_exception()


class MinLengthValidator(Criterion):

    def _validate(self, value, container, field):
        value = str(value)
        if len(value) < self.expression:
            raise self.create_exception()

        return value


class MaxLengthValidator(Criterion):

    def _validate(self, value, container, field):
        value = str(value)
        if len(value) > self.expression:
            raise self.create_exception()

        return value


class MinimumValidator(Criterion):

    def _validate(self, value, container, field):
        try:
            if value < self.expression:
                raise self.create_exception()
        except TypeError:
            raise self.create_exception()

        return value


class MaximumValidator(Criterion):

    def _validate(self, value, container, field):
        try:
            if value > self.expression:
                raise self.create_exception()
        except TypeError:
            raise self.create_exception()

        return value


class PatternValidator(Criterion):

    def _validate(self, value, container, field):
        pattern = re.compile(self.expression)\
            if isinstance(self.expression, str)\
            else self.expression
        if pattern.match(value) is None:
            raise self.create_exception()

        return value


class RequestValidator:
    def __init__(self, fields):
        default = dict(
            required=False,
            form=True,
            query_string=False
        )

        # Merging default specification
        self.fields = {}
        for field_name, specification in fields.items():
            default_copy = default.copy()
            default_copy.update(fields[field_name])
            self.fields[field_name] = Field(field_name, **specification)

    def __call__(self, form=None, query_string=None, *args, **kwargs):
        for field_name, field in self.fields.items():

            if query_string and field.query_string:
                query_string = field.validate(query_string)

            if field.form \
                    and field.title is not None \
                    and (
                        query_string is None or field.title not in query_string
                    ):
                form = field.validate(form)

        return form, query_string


def validate(**fields):

    def decorator(func):
        validator = RequestValidator(fields)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            validator(form=context.form, query_string=context.query_string)
            return func(*args, **kwargs)

        return wrapper

    return decorator
